from logging.config import dictConfig

import pytest
from django.test import Client

from boston_logger.context_managers import (
    RequestDirection,
    RequestEdge,
    SensitivePathContext,
)
from boston_logger.sensitive_paths import (
    MASK_STRING,
    SensitivePaths,
    add_mask_processor,
    remove_mask_processor,
)

dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "loggers": {"": {"level": "INFO", "propagate": True}},
    }
)


def test_middleware_disabled(caplog, mocker):
    mocker.patch("boston_logger.config.config.ENABLE_LOGGING_MIDDLEWARE", False)
    c = Client()
    c.get("/")
    assert len(caplog.records) == 0


def test_middleware_blocklist(caplog, mocker):
    mocker.patch("boston_logger.config.config.MIDDLEWARE_BLOCKLIST", ["index"])
    c = Client()
    c.get("/")
    # Blocking index, actually blocks everything, because BLOCKLIST entries are
    # reversed in the url conf, and then checked as prefixes
    c.get("/log_no_resp_data")
    assert len(caplog.records) == 0


@pytest.mark.parametrize(
    "LOG_RESPONSE_CONTENT, expected_response_data",
    [
        (
            False,
            {"status_code": 200},
        ),
        (
            True,
            {
                "status_code": 200,
                "data": {"obj1": {"key1": MASK_STRING}},
            },
        ),
    ],
)
def test_incoming_request(caplog, mocker, LOG_RESPONSE_CONTENT, expected_response_data):
    mocker.patch("boston_logger.config.config.ENABLE_SENSITIVE_PATHS_PROCESSOR", True)

    mocker.patch(
        "boston_logger.config.config.LOG_RESPONSE_CONTENT", LOG_RESPONSE_CONTENT
    )

    add_mask_processor("Pat1", SensitivePaths("obj1/key1"))

    with SensitivePathContext("Pat1"):
        c = Client()
        c.get("/")

    remove_mask_processor("Pat1")

    assert len(caplog.records) == 2

    expected_start_record = {
        "direction": RequestDirection.INCOMING,
        "edge": RequestEdge.START,
        "end_time": None,
        "exc_info": None,
        "exc_text": None,
        "filename": "context_managers.py",
        "funcName": "log_incoming_request_event",
        "levelname": "INFO",
        "levelno": 20,
        "message": "INCOMING (start): GET /",
        "msg": "INCOMING (start): GET /",
        "name": "boston_logger",
        "notes": None,
        "response": {},
        "response_time_ms": -1,
        "smart": True,
        "stack_info": None,
    }
    # all of expected_start_record is contained in the start log record
    assert expected_start_record.items() <= caplog.records[0].__dict__.items()
    # Can't predict the pathname in all environments
    assert "pathname" in caplog.records[0].__dict__

    expected_start_request = {
        "data": {"raw_body": ""},
        "headers": {"HTTP_COOKIE": ""},
        "method": "GET",
        "path": "/",
        "remote_addr": "127.0.0.1",
        "url_scheme": "http",
    }
    assert expected_start_request.items() <= caplog.records[0].request.items()

    expected_end_record = {
        "direction": RequestDirection.INCOMING,
        "edge": RequestEdge.END,
        "exc_info": None,
        "exc_text": None,
        "filename": "context_managers.py",
        "funcName": "log_incoming_request_event",
        "levelname": "INFO",
        "levelno": 20,
        "message": "INCOMING (end): GET / (200)",
        "module": "context_managers",
        "msg": "INCOMING (end): GET / (200)",
        "name": "boston_logger",
        "notes": None,
        "smart": True,
        "stack_info": None,
    }
    # all of expected_end_record is contained in the end log record
    assert expected_end_record.items() <= caplog.records[1].__dict__.items()

    assert expected_response_data == caplog.records[1].response


@pytest.mark.parametrize(
    "LOG_RESPONSE_CONTENT",
    [
        False,
        True,
    ],
)
def test_incoming_request_no_data(caplog, mocker, LOG_RESPONSE_CONTENT):
    mocker.patch(
        "boston_logger.config.config.LOG_RESPONSE_CONTENT", LOG_RESPONSE_CONTENT
    )

    c = Client()
    c.get("/log_no_resp_data")

    expected_response_data = {
        "data": {"NOT_LOGGED": "_log_data == False"},
        "status_code": 200,
    }
    assert expected_response_data == caplog.records[1].response



INSTALLED_APPS = ("testapp",)

SECRET_KEY = "0VHAq7X9g5QCIoym4WTTarYEIhyXAu4H0rxcsu2FqzIjx4BOJXsdlxchKTpJkCCK"

DEBUG = True

USE_TZ = True

ROOT_URLCONF = "testapp.testapp.urls"

MIDDLEWARE = ["boston_logger.middleware.RequestResponseLoggerMiddleware"]



from django.http import JsonResponse


def index(request):
    return JsonResponse({"obj1": {"key1": "value"}})


def log_no_resp_data(request):
    resp = JsonResponse({"obj1": {"key1": "value"}})
    resp._log_data = False
    return resp



from datetime import datetime
from time import sleep
from unittest.mock import MagicMock

import pytest
import requests

from boston_logger.context_managers import (
    RequestDirection,
    RequestEdge,
    RequestLogContext,
    SensitivePathContext,
    SensitivePathRequestContext,
    log_incoming_request_event,
    log_outgoing_request_event,
)
from boston_logger.sensitive_paths import (
    MASK_STRING,
    SensitivePaths,
    add_mask_processor,
    remove_mask_processor,
    sanitize_data,
    sanitize_querystring,
    sanitize_url,
)

payload = {
    "key1": "value1",
    "key2": "value2",
    "key3": "value3",
}


class Fake(object):
    pass


class ExceptionContext:
    def __init__(self):
        self.exc_info = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exc_info = (exc_type, exc_val, exc_tb)
        return True


class TestSensitivePathContext:
    @pytest.fixture(autouse=True, scope="class")
    def setup(self, class_mocker):
        class_mocker.patch(
            "boston_logger.config.config.ENABLE_SENSITIVE_PATHS_PROCESSOR", True
        )
        add_mask_processor("Pat1", SensitivePaths("key1"))
        add_mask_processor("Pat2", SensitivePaths("key2"))
        add_mask_processor("Pat3", SensitivePaths("key3"))
        yield
        remove_mask_processor("Pat1")
        remove_mask_processor("Pat2")
        remove_mask_processor("Pat3")

    def test_request_post_body(self):
        logger = MagicMock()
        start = datetime.now()
        end = datetime.now()

        s = requests.Session()
        request = s.prepare_request(
            requests.Request("POST", "https://example.com", data={"key1": "hide"})
        )

        with SensitivePathContext("Pat1"):
            log_outgoing_request_event(
                start=start,
                end=end,
                logger=logger,
                request=request,
            )

        # First positional is the message
        assert (
            logger.info.call_args[0][0] == "OUTGOING (end): POST https://example.com/"
        )
        # kwargs extra meets expectations
        extra = logger.info.call_args[1]["extra"]
        # Masked and form encoded
        assert extra["request"]["data"] == "key1=%2A%2A%2A+masked+%2A%2A%2A"

    def test_request_post_body_json(self):
        logger = MagicMock()
        start = datetime.now()
        end = datetime.now()

        s = requests.Session()
        request = s.prepare_request(
            requests.Request(
                "POST",
                "https://example.com",
                json={"key1": "hide"},
            )
        )

        with SensitivePathContext("Pat1"):
            log_outgoing_request_event(
                start=start,
                end=end,
                logger=logger,
                request=request,
            )

        # First positional is the message
        assert (
            logger.info.call_args[0][0] == "OUTGOING (end): POST https://example.com/"
        )
        # kwargs extra meets expectations
        extra = logger.info.call_args[1]["extra"]
        # Masked and form encoded
        assert extra["request"]["data"]["key1"] == MASK_STRING

    def test_path_context_single(self):
        with SensitivePathContext("Pat1"):
            sanitized = sanitize_data(payload)
            assert sanitized["key1"] == MASK_STRING
            assert sanitized["key2"] != MASK_STRING
            assert sanitized["key3"] != MASK_STRING

    def test_path_context_list(self):
        with SensitivePathContext(["Pat1", "Pat2"]):
            sanitized = sanitize_data(payload)
            assert sanitized["key1"] == MASK_STRING
            assert sanitized["key2"] == MASK_STRING
            assert sanitized["key3"] != MASK_STRING

    def test_path_context_nested(self):
        with SensitivePathContext("Pat1"):
            sanitized = sanitize_data(payload)
            assert sanitized["key1"] == MASK_STRING
            assert sanitized["key2"] != MASK_STRING
            assert sanitized["key3"] != MASK_STRING
            with SensitivePathContext("Pat2"):
                sanitized = sanitize_data(payload)
                assert sanitized["key1"] == MASK_STRING
                assert sanitized["key2"] == MASK_STRING
                assert sanitized["key3"] != MASK_STRING

    def test_sanitize_request_data_dict(self):
        request = Fake()
        request._apply_mask_processors = "Pat1"
        with SensitivePathRequestContext(request):
            sanitized = sanitize_data(payload)
        assert sanitized["key1"] == MASK_STRING
        assert sanitized["key2"] != MASK_STRING
        assert sanitized["key3"] != MASK_STRING

    def test_sanitize_request_data_str_no_parse_no_mask(self, mocker):
        mocker.patch("boston_logger.config.config.PREFER_TEXT_FALLBACK_MASKING", False)
        request = Fake()
        data = "not a query string"
        with SensitivePathRequestContext(request):
            sanitized = sanitize_data(data)
        assert sanitized == data

    def test_sanitize_request_data_str_no_parse_mask(self, mocker):
        mocker.patch("boston_logger.config.config.PREFER_TEXT_FALLBACK_MASKING", True)
        request = Fake()
        data = "not a query string"
        with SensitivePathRequestContext(request):
            sanitized = sanitize_querystring(data)
        assert sanitized == MASK_STRING

    def test_sanitize_request_data_str_parse(self):
        request = Fake()
        request._apply_mask_processors = "Pat1"
        data = "key1=hide&key2=show"
        with SensitivePathRequestContext(request):
            sanitized = sanitize_querystring(data)
        # Masked and encoded
        assert sanitized == "key1=%2A%2A%2A+masked+%2A%2A%2A&key2=show"

    def test_sanitize_request_data_url_parse(self):
        request = Fake()
        request._apply_mask_processors = "Pat1"
        data = "http://example.com/?key1=hide&key2=show"
        with SensitivePathRequestContext(request):
            sanitized = sanitize_url(data)
        # Masked and encoded
        assert (
            sanitized == "http://example.com/?key1=%2A%2A%2A+masked+%2A%2A%2A&key2=show"
        )

    def test_sanitize_request_data_url_no_parse(self, mocker):
        mocker.patch("boston_logger.config.config.PREFER_TEXT_FALLBACK_MASKING", True)
        # https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlparse
        # Unmatched square brackets in the netloc attribute will raise a ValueError.
        # Included for coverage of failed URL parsing
        data = "http://exa[mple.com/?key1=hide&key2=show"
        sanitized = sanitize_url(data)
        assert sanitized == MASK_STRING

    def test_sanitize_request_data_absurd(self):
        request = Fake()
        with SensitivePathRequestContext(request):
            sanitized = sanitize_data([])
        # Not a dict or str, just gets returned
        assert sanitized == []


@pytest.mark.parametrize(
    "test_func",
    [
        log_incoming_request_event,
        log_outgoing_request_event,
    ],
)
class TestRequestLogContext:
    def test_defaults(self, test_func):
        # For coverage of config logger lookup
        with RequestLogContext(make_log_func=test_func):
            pass

    def test_logger(self, test_func):
        logger = MagicMock()

        with RequestLogContext(
            logger=logger,
            make_log_func=test_func,
        ):
            pass

        # Log start and end
        assert logger.info.call_count == 2


class Test_log_request_event:
    @pytest.mark.parametrize(
        "test_func",
        [
            log_incoming_request_event,
            log_outgoing_request_event,
        ],
    )
    def test_info(self, test_func):
        logger = MagicMock()
        start = datetime.now()
        sleep(0.01)
        end = datetime.now()
        delta = int((end - start).total_seconds() * 1000)

        test_func(
            start=start,
            end=end,
            logger=logger,
        )

        logger.info.assert_called_once()
        assert logger.info.call_args[1]["extra"]["response_time_ms"] == delta
        logger.error.assert_not_called()

    @pytest.mark.parametrize(
        "test_func",
        [
            log_incoming_request_event,
            log_outgoing_request_event,
        ],
    )
    def test_error(self, test_func):
        logger = MagicMock()
        start = datetime.now()
        end = datetime.now()

        with ExceptionContext() as context:
            raise ValueError("I don't like it.")

        # WHEN logging an Exception
        test_func(
            start=start,
            end=end,
            logger=logger,
            exc_info=context.exc_info,
        )

        # THEN exc_info gets passed to logger.error
        logger.error.assert_called_once()
        assert logger.error.call_args[1]["exc_info"] == context.exc_info
        # AND logger.info is not called
        logger.info.assert_not_called()

    def test_incoming_start(self):
        logger = MagicMock()
        start = datetime.now()
        end = datetime.now()
        edge = RequestEdge.START
        direction = RequestDirection.INCOMING

        request = Fake()
        request.method = ("GET",)
        request.META = {"REMOTE_ADDR": "127.0.0.1", "HTTP_HEADER": "orly"}
        request.scheme = "https"
        request.path = "/index/"
        request.POST = {}
        request.GET = {}

        # WHEN logging an INCOMING request
        log_incoming_request_event(
            start=start,
            end=end,
            logger=logger,
            request=request,
            edge=edge,
        )

        # First positional is the message
        assert logger.info.call_args[0][0] == "INCOMING (start): ('GET',) /index/"
        # kwargs extra meets expectations
        extra = logger.info.call_args[1]["extra"]
        assert extra["direction"] == direction
        assert extra["edge"] == edge

    def test_incoming_end(self):
        logger = MagicMock()
        start = datetime.now()
        end = datetime.now()
        edge = RequestEdge.END
        direction = RequestDirection.INCOMING

        request = Fake()
        request.method = ("GET",)
        request.body = "Some Payload"
        request.META = {"REMOTE_ADDR": "127.0.0.1", "HTTP_HEADER": "orly"}
        request.scheme = "https"
        request.path = "/index/"
        request.POST = {}
        request.GET = {}

        # WHEN logging an INCOMING request
        log_incoming_request_event(
            start=start,
            end=end,
            logger=logger,
            request=request,
            edge=edge,
        )

        # First positional is the message
        assert logger.info.call_args[0][0] == "INCOMING (end): ('GET',) /index/"
        # kwargs extra meets expectations
        extra = logger.info.call_args[1]["extra"]
        assert extra["direction"] == direction
        assert extra["edge"] == edge



import json
from datetime import date, datetime
from decimal import Decimal
from logging import Filter, Formatter


class RequestEdgeEndFilter(Filter):
    def filter(self, record):
        from .context_managers import RequestEdge

        if not getattr(record, "smart", False):
            # non-smart logs always get recorded
            return True

        # Smart logs only get recorded on END
        return record.edge == RequestEdge.END


class ObjectTypeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return {
                # sort values to make reading logs easier as a user
                "value": sorted(obj),
                "type": "set",
            }

        if isinstance(obj, datetime):
            return {
                "value": str(obj),
                "type": "datetime",
            }

        if isinstance(obj, date):
            return {
                "value": str(obj),
                "type": "date",
            }

        if isinstance(obj, Decimal):
            return {
                "value": str(obj),
                "type": "Decimal",
            }

        return super().default(obj)


class JsonFormatter(Formatter):
    def __init__(self, *args, **kwargs):
        self.default_extra = kwargs.pop("default_extra", {})
        super().__init__(*args, **kwargs)

    def format(self, record):
        from .config import config

        # Normal tracing stuff
        log_data = {
            "timestamp": self.formatTime(record),
            **self.default_extra,
            "filename": record.filename,
            "funcName": record.funcName,
            "levelname": record.levelname,
            "levelno": record.levelno,
            "lineno": record.lineno,
            "module": record.module,
            "loggerName": record.name,
        }

        if getattr(record, "smart", False):
            log_data.update(
                {
                    # Custom Fields
                    "start_time": getattr(record, "start_time", ""),
                    "end_time": getattr(record, "end_time", ""),
                    "response_time_ms": getattr(record, "response_time_ms", ""),
                    "request": record.request,
                    "response": getattr(record, "response", None),
                    "notes": getattr(record, "notes", None),
                }
            )

        # Always fields
        log_data.update(
            {
                "msg": super().format(record),
                **getattr(record, "extra", {}),
            }
        )

        resp = json.dumps(log_data, cls=ObjectTypeEncoder)
        if config.MAX_JSON_DATA_TO_LOG and len(resp) > config.MAX_JSON_DATA_TO_LOG:
            log_data["max_data_exceeded"] = True
            truncate_length = config.MAX_JSON_DATA_TO_LOG - 50
            response_obj = log_data.get("response")
            if response_obj and truncate_length > 0:
                response_data = str(response_obj.get("data", ""))
                if len(response_data) > truncate_length:
                    # Update mutatable resonse_obj inside mutatable log_data
                    response_obj["data"] = (
                        response_data[:truncate_length] + " **TRUNCATED**"
                    )

            resp = json.dumps(log_data, cls=ObjectTypeEncoder)

        return resp


class SmartFormatter(Formatter):
    def limited_size_repr(self, data, length):
        data = repr(data)
        if len(data) > length:
            data = data[:length] + "..."
        return data

    def format(self, record):
        from .context_managers import RequestDirection, RequestEdge

        log_msg = [super().format(record)]

        if not getattr(record, "smart", False):
            return log_msg[0]

        if (
            record.edge == RequestEdge.START
            and record.direction == RequestDirection.OUTGOING
        ):
            pass
        else:
            from .config import config

            req = record.request

            data = req.get("data")
            if data is not None:
                data = self.limited_size_repr(data, config.MAX_VERBOSE_OUTPUT_LENGTH)
                log_msg.append(f"  Request Data: {data}")

            headers = req.get("headers")
            if headers is not None:
                headers = self.limited_size_repr(
                    headers, config.MAX_VERBOSE_OUTPUT_LENGTH
                )
                log_msg.append(f"  Request Headers: {headers}")

            resp = getattr(record, "response", None)
            if resp is not None:
                data = resp.get("data")
                if data is None:
                    data = "(empty)"
                else:
                    data = self.limited_size_repr(
                        data, config.MAX_VERBOSE_OUTPUT_LENGTH
                    )
                log_msg.append(f"  Response Data: {data}")

            log_msg.append("\n")

        return "\n".join(log_msg)



import os
from logging.config import dictConfig

import pytest
import requests

from boston_logger.config import config

# from boston_logger import requests_monkey_patch  # NOQA
from boston_logger.context_managers import (
    RequestDirection,
    RequestEdge,
    SensitivePathContext,
)
from boston_logger.sensitive_paths import (
    MASK_STRING,
    SensitivePaths,
    add_mask_processor,
    remove_mask_processor,
)

dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "loggers": {"": {"level": "INFO", "propagate": True}},
    }
)


@pytest.fixture(autouse=True, scope="module")
def setup(module_mocker):
    module_mocker.patch(
        "boston_logger.config.config.ENABLE_SENSITIVE_PATHS_PROCESSOR", True
    )

    # Test the reconfigure feature triggering monkey patch
    orig_request = requests.request
    assert config.ENABLE_REQUESTS_LOGGING is False
    os.environ["BOSTON_LOGGER_ENABLE_REQUESTS_LOGGING"] = "yes"
    config.reconfigure()
    assert config.request_logging_enabled is True
    assert requests.request is not orig_request
    yield
    os.environ.pop("BOSTON_LOGGER_ENABLE_REQUESTS_LOGGING")
    config.reconfigure()
    # The monkey patch can not be undone, use `ENABLE_OUTBOUND_REQUEST_LOGGING`
    # To avoid logging after patching
    assert requests.request is not orig_request


def test_csv_api(caplog):
    requests.get("https://ip4.me/api/")
    assert "OUTGOING (start): GET https://ip4.me/api/" in caplog.text
    assert "OUTGOING (end): GET https://ip4.me/api/ (200)" in caplog.text


def test_csv_api_nolog(caplog, mocker):
    mocker.patch("boston_logger.config.config.ENABLE_OUTBOUND_REQUEST_LOGGING", False)
    requests.get("https://ip4.me/api/")
    assert len(caplog.records) == 0


def test_json_api(caplog):
    add_mask_processor("Facts", SensitivePaths("fact"))

    with SensitivePathContext("Facts"):
        requests.get("https://catfact.ninja/fact")

    remove_mask_processor("Facts")

    assert "OUTGOING (start): GET https://catfact.ninja/fact" in caplog.text
    assert "OUTGOING (end): GET https://catfact.ninja/fact (200)" in caplog.text
    # Two logs happened
    assert len(caplog.records) == 2

    expected_start_record = {
        "direction": RequestDirection.OUTGOING,
        "edge": RequestEdge.START,
        "end_time": None,
        "exc_info": None,
        "exc_text": None,
        "filename": "context_managers.py",
        "funcName": "log_outgoing_request_event",
        "levelname": "INFO",
        "message": "OUTGOING (start): GET https://catfact.ninja/fact",
        "module": "context_managers",
        "msg": "OUTGOING (start): GET https://catfact.ninja/fact",
        "name": "boston_logger",
        "notes": None,
        "request": {"method": "GET", "url": "https://catfact.ninja/fact"},
        "response": {},
        "response_time_ms": -1,
        "smart": True,
        "stack_info": None,
    }
    # all of expected_start_record is contained in the start log record
    assert expected_start_record.items() <= caplog.records[0].__dict__.items()

    expected_end_record = {
        "direction": RequestDirection.OUTGOING,
        "edge": RequestEdge.END,
        "exc_info": None,
        "exc_text": None,
        "filename": "context_managers.py",
        "funcName": "log_outgoing_request_event",
        "levelname": "INFO",
        "message": "OUTGOING (end): GET https://catfact.ninja/fact (200)",
        "module": "context_managers",
        "msg": "OUTGOING (end): GET https://catfact.ninja/fact (200)",
        "name": "boston_logger",
        "notes": None,
        "smart": True,
        "stack_info": None,
    }
    # all of expected_end_record is contained in the end log record
    assert expected_end_record.items() <= caplog.records[1].__dict__.items()

    expected_end_request = {
        "method": "GET",
        "path": "/fact",
        "url": "https://catfact.ninja/fact",
    }
    assert expected_end_request.items() <= caplog.records[1].request.items()

    # Masking works
    assert caplog.records[1].response["data"]["fact"] == MASK_STRING



import copy

import pytest

from boston_logger.sensitive_paths import (
    MASK_STRING,
    SensitivePaths,
    _global_masks,
    _mask_processors,
    add_mask_processor,
    chain_mask,
    remove_mask_processor,
    sanitize_data,
)


def test_chain_mask_handles_none():
    # Mostly for coverage
    assert chain_mask(None) is None


def test_remove_safe():
    # GIVEN processor that does not exist
    assert "N/A" not in _mask_processors
    assert "N/A" not in _global_masks
    # WHEN removed
    remove_mask_processor("N/A")
    # THEN there is no error


@pytest.mark.parametrize(
    "paths, expected_root_paths",
    [
        (
            # Leading and Trailing /'s removed
            # Resulting duplicates, collapsed
            ["/obj1/key1", "obj1/key1", "obj1/key2/"],
            {
                "obj1": {
                    "key1": True,
                    "key2": True,
                },
            },
        ),
        (
            # Midpath *'s honored
            # Terminal *'s removed
            ["/obj1/*/key1", "obj1/key2/*"],
            {
                "obj1": {
                    "*": {
                        "key1": True,
                    },
                    "key2": True,
                },
            },
        ),
        (
            # Depth first collapsed
            ["obj1/nested/key1", "obj1/nested"],
            {
                "obj1": {
                    "nested": True,
                },
            },
        ),
        (
            # Deeper Depth first collapsed
            ["obj1/nested/key1/deeper", "obj1/nested/key1", "obj1/nested"],
            {
                "obj1": {
                    "nested": True,
                },
            },
        ),
        (
            # Depth second collapsed
            ["obj1/nested", "obj1/nested/key1"],
            {
                "obj1": {
                    "nested": True,
                },
            },
        ),
        (
            # Deeper Depth second collapsed
            ["obj1/nested", "obj1/nested/key1", "obj1/nested/key1/deeper"],
            {
                "obj1": {
                    "nested": True,
                },
            },
        ),
    ],
)
def test_root_paths(paths, expected_root_paths):
    """
    GIVEN a set of paths
    THEN an expected root_paths dict will be created
    """

    sp = SensitivePaths(*paths)

    assert sp.root_paths == expected_root_paths


class TestPaths:
    @pytest.fixture(autouse=True)
    def setup(self, mocker):
        mocker.patch(
            "boston_logger.config.config.ENABLE_SENSITIVE_PATHS_PROCESSOR", True
        )
        self.sp = SensitivePaths(
            "obj1/key1", "obj1/key2", "obj2/*/wild", "obj2/nested/calm"
        )

    @pytest.mark.parametrize(
        "data, expected_masked_data",
        [
            (
                # Simple match
                # Extra data passed
                {
                    "obj2": "shown",
                    "obj1": {
                        "key1": "secret",
                        "key2": "hidden",
                        "key3": "shown",
                    },
                },
                {
                    "obj1": {
                        "key1": MASK_STRING,
                        "key2": MASK_STRING,
                        "key3": "shown",
                    },
                    "obj2": "shown",
                },
            ),
            (
                # list value
                {"obj1": {"key1": ["x", "y", "z"]}},
                {"obj1": {"key1": [MASK_STRING] * 3}},
            ),
            (
                # lists of objects check each object
                {
                    "obj1": [
                        {
                            "key1": "secret",
                            "key2": "hidden",
                            "key3": "shown",
                        },
                        {
                            "key1": "secret",
                            "key2": "hidden",
                            "key3": "shown",
                        },
                    ],
                },
                {
                    "obj1": [
                        {
                            "key1": MASK_STRING,
                            "key2": MASK_STRING,
                            "key3": "shown",
                        },
                        {
                            "key1": MASK_STRING,
                            "key2": MASK_STRING,
                            "key3": "shown",
                        },
                    ],
                },
            ),
            (
                # deeper lists ignored
                # Extra data passed
                {
                    "obj2": "shown",
                    "obj1": [
                        {
                            "key1": "secret",
                            "key2": "hidden",
                            "key3": "shown",
                        },
                        [
                            {
                                "key1": "secret",
                                "key2": "hidden",
                                "key3": "shown",
                            },
                        ],
                    ],
                },
                {
                    "obj1": [
                        {
                            "key1": MASK_STRING,
                            "key2": MASK_STRING,
                            "key3": "shown",
                        },
                        [
                            {
                                "key1": MASK_STRING,
                                "key2": MASK_STRING,
                                "key3": "shown",
                            },
                        ],
                    ],
                    "obj2": "shown",
                },
            ),
            (
                # Mask past one star
                # But not two level
                {
                    "obj2": {
                        "key1": {
                            "wild": "hidden",
                        },
                        "key2": {
                            "wild": "hidden",
                        },
                        "key3": {
                            "nested": {
                                "wild": "shown",
                            },
                        },
                    },
                },
                {
                    "obj2": {
                        "key1": {
                            "wild": MASK_STRING,
                        },
                        "key2": {
                            "wild": MASK_STRING,
                        },
                        "key3": {
                            "nested": {
                                "wild": "shown",
                            },
                        },
                    },
                },
            ),
            (
                # Lists and stars
                {
                    "obj2": [
                        {
                            "key1": {
                                "wild": "hidden",
                            },
                        },
                        {
                            "key2": {
                                "wild": "hidden",
                            },
                        },
                        {
                            "key3": {
                                "nested": {
                                    "wild": "shown",
                                },
                            },
                        },
                    ]
                },
                {
                    "obj2": [
                        {
                            "key1": {
                                "wild": MASK_STRING,
                            },
                        },
                        {
                            "key2": {
                                "wild": MASK_STRING,
                            },
                        },
                        {
                            "key3": {
                                "nested": {
                                    "wild": "shown",
                                },
                            },
                        },
                    ],
                },
            ),
            (
                # Star and others
                {
                    "obj2": {
                        "key1": {
                            "wild": "hidden",
                        },
                        "nested": {
                            "calm": "hidden",
                        },
                    }
                },
                {
                    "obj2": {
                        "key1": {
                            "wild": MASK_STRING,
                        },
                        "nested": {
                            "calm": MASK_STRING,
                        },
                    }
                },
            ),
        ],
    )
    def test_data(self, data, expected_masked_data):
        # There is no return data
        assert self.sp.process(data) is None
        # Data has been changed in place
        assert data == expected_masked_data


class TestAll:
    @pytest.fixture(autouse=True, scope="class")
    def setup(self, class_mocker):
        class_mocker.patch(
            "boston_logger.config.config.ENABLE_SENSITIVE_PATHS_PROCESSOR", True
        )

    @pytest.mark.parametrize(
        "data, expected_hide_nested, expected_show_nested",
        [
            (
                # dict masked
                {
                    "obj2": "shown",
                    "obj1": {
                        "key1": "secret",
                        "key2": "hidden",
                        "key3": "shown",
                    },
                },
                {
                    MASK_STRING: MASK_STRING,
                },
                {
                    "obj2": MASK_STRING,
                    "obj1": MASK_STRING,
                },
            ),
            (
                # dict with list masked
                {
                    "obj2": "shown",
                    "list1": [
                        "secret",
                        "hidden",
                        "shown",
                    ],
                },
                {
                    MASK_STRING: MASK_STRING,
                },
                {
                    "obj2": MASK_STRING,
                    "list1": MASK_STRING,
                },
            ),
        ],
    )
    def test_data(self, mocker, data, expected_hide_nested, expected_show_nested):
        orig_data = copy.deepcopy(data)
        # Returned data is sanitized
        # ALL is a built-in filter, but not globally active
        assert sanitize_data(data, "ALL") == expected_hide_nested
        # Data has not been altered
        assert data == orig_data

        mocker.patch(
            "boston_logger.config.config.SHOW_NESTED_KEYS_IN_SENSITIVE_PATHS", True
        )
        assert sanitize_data(data, "ALL") == expected_show_nested


class TestGlobal:
    @pytest.fixture(autouse=True, scope="class")
    def setup(self, class_mocker):
        class_mocker.patch(
            "boston_logger.config.config.ENABLE_SENSITIVE_PATHS_PROCESSOR", True
        )
        add_mask_processor(
            "Simple", SensitivePaths("obj1/key1", "list1"), is_global=True
        )
        yield
        remove_mask_processor("Simple")

    @pytest.mark.parametrize(
        "data, expected_hide_nested, expected_show_nested",
        [
            (
                # dict masked
                {
                    "obj2": "shown",
                    "obj1": {
                        "key1": "secret",
                        "key3": "shown",
                    },
                },
                {
                    "obj2": "shown",
                    "obj1": {
                        "key1": MASK_STRING,
                        "key3": "shown",
                    },
                },
                {
                    "obj2": "shown",
                    "obj1": {
                        "key1": MASK_STRING,
                        "key3": "shown",
                    },
                },
            ),
            (
                # nested dict masked
                {
                    "obj2": "shown",
                    "obj1": {
                        "key1": {"maybe_show": "secret"},
                        "key3": "shown",
                    },
                },
                {
                    "obj2": "shown",
                    "obj1": {
                        "key1": {MASK_STRING: MASK_STRING},
                        "key3": "shown",
                    },
                },
                {
                    "obj2": "shown",
                    "obj1": {
                        "key1": {"maybe_show": MASK_STRING},
                        "key3": "shown",
                    },
                },
            ),
            (
                # dict with list masked
                {
                    "obj2": "shown",
                    "list1": [
                        "secret",
                        "hidden",
                        "shown",
                    ],
                },
                {
                    "obj2": "shown",
                    "list1": [MASK_STRING] * 3,
                },
                {
                    "obj2": "shown",
                    "list1": [MASK_STRING] * 3,
                },
            ),
        ],
    )
    def test_data(self, mocker, data, expected_hide_nested, expected_show_nested):
        orig_data = copy.deepcopy(data)
        # Returned data is sanitized by global processor(s)
        assert sanitize_data(data) == expected_hide_nested
        # Data has not been altered
        assert data == orig_data

        # When Show key is enabled, more data is visible
        mocker.patch(
            "boston_logger.config.config.SHOW_NESTED_KEYS_IN_SENSITIVE_PATHS", True
        )
        assert sanitize_data(data) == expected_show_nested



import json
import logging

from boston_logger.context_managers import RequestDirection, RequestEdge
from boston_logger.logger import JsonFormatter, RequestEdgeEndFilter, SmartFormatter

request_edge_filter = RequestEdgeEndFilter()


class Fake(object):
    pass


not_smart_keys = {
    "timestamp",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msg",
    "loggerName",
}


smart_keys = not_smart_keys | {
    "end_time",
    "notes",
    "request",
    "response",
    "response_time_ms",
    "start_time",
}


def test_filter_dumb():
    # All non-smart logs pass
    assert request_edge_filter.filter(Fake()) is True


def test_filter_smart_start():
    record = Fake()
    record.smart = True
    record.edge = RequestEdge.START

    assert request_edge_filter.filter(record) is False


def test_filter_smart_end():
    record = Fake()
    record.smart = True
    record.edge = RequestEdge.END

    # Smart logs only pass on the END edge
    assert request_edge_filter.filter(record) is True


def test_json_smart_format():
    long_msg = "TEST " * 100

    record = logging.makeLogRecord(
        {
            "msg": long_msg,
            "request": long_msg,
            "smart": True,
            "response": {"data": long_msg},
            "extra": {"key": "value"},
        }
    )
    formatter = JsonFormatter()

    fmt = json.loads(formatter.format(record))
    # Smart logs have expected keys, including from extra
    assert set(fmt.keys()) == smart_keys | {"key"}


def test_json_smart_format_default_extra():
    long_msg = "TEST " * 100

    record = logging.makeLogRecord(
        {
            "msg": long_msg,
            "request": long_msg,
            "smart": True,
            "response": {"data": long_msg},
            "extra": {"key": "value"},
        }
    )
    formatter = JsonFormatter(default_extra={"_extra_key": "some value"})

    fmt = json.loads(formatter.format(record))
    # Smart logs have expected keys, including from extra
    assert set(fmt.keys()) == smart_keys | {"key", "_extra_key"}


def test_json_not_smart_format():
    long_msg = "TEST " * 100

    record = logging.makeLogRecord(
        {
            "msg": long_msg,
            "request": long_msg,
            "smart": False,
            "response": {"data": long_msg},
            "extra": {"key": "value"},
        }
    )
    formatter = JsonFormatter()

    fmt = json.loads(formatter.format(record))
    # Non-smart logs only have msg and extra
    assert set(fmt.keys()) == not_smart_keys | {"key"}


def test_json_not_smart_format_default_extra():
    long_msg = "TEST " * 100

    record = logging.makeLogRecord(
        {
            "msg": long_msg,
            "request": long_msg,
            "smart": False,
            "response": {"data": long_msg},
            "extra": {"key": "value"},
        }
    )
    formatter = JsonFormatter(default_extra={"_extra_key": "some value"})

    fmt = json.loads(formatter.format(record))
    # Non-smart logs only have msg, extra, and default_extra
    assert set(fmt.keys()) == not_smart_keys | {"key", "_extra_key"}


def test_smart_format_not_smart():
    long_msg = "TEST " * 100

    record = logging.makeLogRecord(
        {
            "msg": long_msg,
            "request": long_msg,
            "smart": False,
            "response": {"data": long_msg},
            "extra": {"key": "value"},
        }
    )
    formatter = SmartFormatter()

    fmt = formatter.format(record)
    # Not smart is just the msg
    assert fmt == long_msg


def test_smart_format_start_outgoing():
    long_msg = "TEST " * 100

    record = logging.makeLogRecord(
        {
            "msg": long_msg,
            "request": long_msg,
            "smart": True,
            "response": {"data": long_msg},
            "extra": {"key": "value"},
            "edge": RequestEdge.START,
            "direction": RequestDirection.OUTGOING,
        }
    )
    formatter = SmartFormatter()

    fmt = formatter.format(record)
    # Start/Outgoing is just msg
    assert fmt == long_msg



from copy import deepcopy
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

MASK_STRING = "*** masked ***"

_global_masks = set()
_mask_processors = {}


def chain_mask(data):
    """Return data with all values masked."""
    if data is None:
        return None

    if isinstance(data, dict):
        from .config import config

        if config.SHOW_NESTED_KEYS_IN_SENSITIVE_PATHS:
            return {k: MASK_STRING for k in data.keys()}
        else:
            return {MASK_STRING: MASK_STRING}
    elif isinstance(data, list):
        return [chain_mask(x) for x in data]
    else:
        return MASK_STRING


class SensitivePaths:
    def __init__(self, *args: str):
        # str -> Union(dict, bool)
        # Nested dictionaries represent paths
        # A "True" value means the data should be masked from that path and
        # all children.
        self.root_paths = {}

        # '*' matches any key at ONE level.
        # Leading and trailing '/' has no effect
        #
        # args = [ 'obj/key1', '/obj/key2', 'obj/key2/ignored', 'obj/*/nested1/' ]
        # Produces root_paths of:
        #   {
        #       'obj': {
        #           'key1': True,
        #           'key2': True,
        #           '*': {
        #               'nested1': True,
        #           }
        #       }
        #   }

        for path in args:
            current_dict = self.root_paths
            # Leading and trailing '/' has no effect
            keys = path.strip("/").split("/")
            while keys:
                k = keys.pop(0)
                if len(keys) == 1 and keys[0] == "*":
                    # Terminal "*" is the same as not having it
                    # Everything from the current path down is sanitized
                    current_dict[k] = True
                    break
                if len(keys) == 0:
                    current_dict[k] = True
                else:
                    current_dict.setdefault(k, {})
                    current_dict = current_dict[k]
                    if current_dict is True:
                        # We hit a previous path, underwhich everything should
                        # be masked. Skip the rest of this path, it will all be
                        # masked
                        break

    def process(self, data: dict):
        """Mutates data to apply sanitation.

        Sanitation only happens if config.ENABLE_SENSITIVE_PATHS_PROCESSOR is True.
        """
        from .config import config

        if config.ENABLE_SENSITIVE_PATHS_PROCESSOR:
            SensitivePaths._sanitize_dict(self.root_paths, data)

    @staticmethod
    def _sanitize_dict(paths, data: dict):
        """Mutates data to sanitize values in matching paths."""

        if "*" in paths:
            nested_paths = paths["*"]
            if nested_paths is True:
                from .config import config

                if config.SHOW_NESTED_KEYS_IN_SENSITIVE_PATHS:
                    # Special case for single '*' entry
                    data.update(chain_mask(data))
                else:
                    data.clear()
                    data[MASK_STRING] = MASK_STRING
            else:
                # Normally check all values for the paths under the '*'
                for v in data.values():
                    SensitivePaths._sanitize_any(nested_paths, v)

        for k, v in data.items():
            nested_paths = paths.get(k)
            if nested_paths is None:
                # k not in paths -> don't sanitize
                pass
            elif nested_paths is True:
                # Mask below this path
                data[k] = chain_mask(v)
            else:
                # Process nested_paths
                SensitivePaths._sanitize_any(nested_paths, data[k])

    @staticmethod
    def _sanitize_any(paths, data):
        """Mutates data to sanitize values in matching paths."""

        if isinstance(data, dict):
            SensitivePaths._sanitize_dict(paths, data)

        elif isinstance(data, (list, tuple)):
            # If there are Lists of objects process each object.
            # The list does not add a path element
            # SensitivePaths( 'obj1/key1' )
            # WOULD mask the key1's in the folowing object:
            #   {
            #       'obj1': [
            #           {
            #               'key1': 'sensitive',
            #           }, {
            #               'key1': 'sensitive',
            #           },
            #       ]
            #   }

            for item in data:
                SensitivePaths._sanitize_any(paths, item)
        # else:
        # Data isn't itterable, and path didn't match, nothing to do


def add_mask_processor(mask_name, processor, *, is_global=False):
    """Register processor as mask_name

    processor must have a method process(data, is_query_string) which mutates
    data to apply sanitation.

    if is_global == True, this processor will apply to all data sanitation
    """
    _mask_processors[mask_name] = processor
    if is_global:
        _global_masks.add(mask_name)


def remove_mask_processor(mask_name):
    """Unregister a processor, and remove from global list if possible."""
    _mask_processors.pop(mask_name, None)
    try:
        # Remove from global if it exists
        _global_masks.remove(mask_name)
    except KeyError:
        pass


# All is a special name to match all data
add_mask_processor("ALL", SensitivePaths("*"))


def sanitize_data(data: dict, *mask_names):
    """Return copy of data that has been sanitized.

    Global masks, the current SensitivePathContext and any positional args will
    be applied.
    """
    # Avoid circular import
    from .context_managers import SensitivePathContext

    context_mask_names = SensitivePathContext.get_mask_names()
    mask_names = set(mask_names) | context_mask_names | _global_masks

    masked_data = deepcopy(data)

    for mask_name in mask_names:
        _mask_processors[mask_name].process(masked_data)

    return masked_data


def sanitize_url(url: str, *mask_names):
    """Return copy of url with query string that has been sanitized.

    Global masks, the current SensitivePathContext and any positional args will
    be applied.
    """

    try:
        parts = urlparse(url)
        return urlunparse(parts._replace(query=sanitize_querystring(parts.query)))
    except ValueError:
        from .config import config

        return MASK_STRING if config.PREFER_TEXT_FALLBACK_MASKING else url


def sanitize_querystring(data: str, *mask_names):
    """Return copy of data sanitized as if it was application/x-www-form-urlencoded data

    Global masks, the current SensitivePathContext and any positional args will
    be applied.
    """

    try:
        masked_query_dict = sanitize_data(
            parse_qs(data, keep_blank_values=True, strict_parsing=True), *mask_names
        )
        return urlencode(masked_query_dict, doseq=True)
    except ValueError:
        from .config import config

        return MASK_STRING if config.PREFER_TEXT_FALLBACK_MASKING else data



"""Middleware for request/response logging."""

import json

from django.urls import NoReverseMatch, reverse

from .config import config
from .context_managers import (
    RequestDirection,
    RequestLogContext,
    log_incoming_request_event,
)
from .sensitive_paths import sanitize_querystring


class RequestResponseLoggerMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if not config.ENABLE_LOGGING_MIDDLEWARE:
            return self.get_response(request)

        if self._should_ignore(request):
            return self.get_response(request)

        # Store request content now because it'll be too late to access later on.
        request.body

        if request.FILES:
            # TODO - when there's files, there's nothing else to add to req_data?
            req_data = {
                "file_list": [
                    request.FILES[filename].name
                    for filename, _ in request.FILES.items()
                ]
            }
        else:
            req_data = request.body.decode("utf-8", "replace")
            try:
                # TODO - no attempt at using content negotiation headers
                # this will fail on form encoded POSTs, yeah?
                req_data = json.loads(req_data)
            except json.JSONDecodeError:
                mask_names = getattr(request, "_apply_mask_processors", [])
                req_data = {"raw_body": sanitize_querystring(req_data, *mask_names)}

        with RequestLogContext(
            request=request,
            request_data=req_data,
            direction=RequestDirection.INCOMING,
            make_log_func=log_incoming_request_event,
        ) as log_context:
            # Could be django.http.response.HttpResponse
            response = self.get_response(request)
            log_context.notes = getattr(request, "_request_notes", None)
            log_context.response = response

        return response

    @staticmethod
    def _should_ignore(request):
        """
        Checks if the requested url should logged or not
        By default it will ignore admin and swagger requests
        """
        for url_name in config.MIDDLEWARE_BLOCKLIST:
            try:
                if request.path.startswith(reverse(url_name)):
                    return True
            except NoReverseMatch:
                pass
        return False



import requests

original_request_method = requests.request


def logged_request(method, url, **kwargs):
    from .config import config

    notes = kwargs.pop("notes") if "notes" in kwargs else None

    if not config.ENABLE_OUTBOUND_REQUEST_LOGGING:
        return original_request_method(method, url, **kwargs)

    # TODO allow/block lists here for remote URLs

    from .context_managers import RequestLogContext, log_outgoing_request_event

    with RequestLogContext(
        method=method,
        url=url,
        make_log_func=log_outgoing_request_event,
    ) as log_context:
        response = original_request_method(method, url, **kwargs)
        request = response.request

        log_context.request = request
        log_context.response = response
        log_context.notes = notes

    return response


requests.api.request = logged_request
requests.request = logged_request



"""Context manager to wrap requests."""

import itertools
import json
import logging
from datetime import datetime
from enum import Enum

from .sensitive_paths import sanitize_data, sanitize_querystring, sanitize_url

RequestDirection = Enum("RequestDirection", "INCOMING OUTGOING")
RequestEdge = Enum("RequestEdge", "START END")


# Probably not thread safe
class SensitivePathContext:
    _mask_names = []  # List[Set]

    def __init__(self, paths):
        # You probably didn't mean to pass a single name, but you can
        if isinstance(paths, str):
            paths = set([paths])
        SensitivePathContext._mask_names.append(set(paths))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        SensitivePathContext._mask_names.pop(-1)

    @classmethod
    def get_mask_names(cls):
        return set(itertools.chain(*cls._mask_names))


class SensitivePathRequestContext(SensitivePathContext):
    def __init__(self, request):
        # Should be a list or set of named mask processors already
        # registered with sensitive_paths.add_mask_processor
        # Or name of a single processor
        mask_names = getattr(request, "_apply_mask_processors", [])
        super().__init__(mask_names)


class RequestLogContext:
    def __init__(
        self,
        *,
        logger=None,
        request=None,
        request_data=None,
        notes=None,
        msg=None,
        method=None,
        url=None,
        direction=RequestDirection.OUTGOING,
        make_log_func,
    ):
        if logger is None:
            # Avoid importing config which might hit django/constance in global scope
            from .config import config

            self.logger = logging.getLogger(config.LOGGER_NAME)
        else:
            self.logger = logger

        self.start_time = datetime.now()
        self.response = None
        self.request = request
        self.request_data = request_data
        self.notes = notes
        self.msg = msg
        self.method = method
        self.url = url
        self.direction = direction
        self.make_log_func = make_log_func

    def __enter__(self):
        self.make_log_func(
            start=self.start_time,
            end=None,
            logger=self.logger,
            request=self.request,
            request_data=self.request_data,
            notes=self.notes,
            msg=self.msg,
            method=self.method,
            url=self.url,
            direction=self.direction,
            edge=RequestEdge.START,
        )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = datetime.now()

        self.make_log_func(
            start=self.start_time,
            end=self.end_time,
            logger=self.logger,
            request=self.request,
            response=self.response,
            request_data=self.request_data,
            exc_info=(exc_type, exc_val, exc_tb) if exc_type else None,
            notes=self.notes,
            msg=self.msg,
            direction=self.direction,
            edge=RequestEdge.END,
        )


def calculate_response_time_ms(start_time: datetime, end_time: datetime):
    if not end_time:
        return -1

    return int((end_time - start_time).total_seconds() * 1000)


# Expects to handle only requests lib Requests
def log_outgoing_request_event(
    *,
    start,
    end,
    logger,
    request=None,
    response=None,
    request_data=None,
    exc_info=None,
    notes=None,
    msg=None,
    method=None,
    url=None,
    direction=None,  # only exists to match existing interface
    edge=RequestEdge.END,
):
    request_info = {}
    response_info = {}

    log_msg = None
    direction = RequestDirection.OUTGOING

    with SensitivePathRequestContext(request):
        url = sanitize_url(url)
        if edge == RequestEdge.START:
            # We can't get access to the prepared request on the START edge
            # The monkey patch will provide us some information.
            method = (method or "").upper()
            request_info = {
                "method": method,
                "url": url,
            }

            log_msg = f"OUTGOING (start): {method} {url}"

        else:
            # assert edge == RequestEdge.END
            if request:
                url = sanitize_url(request.url)
                request_info = {
                    "method": request.method,
                    "url": url,
                    "path": sanitize_url(request.path_url),
                    "headers": sanitize_data(dict(request.headers)),
                }

                if request.body:
                    if isinstance(request.body, bytes):
                        body = request.body.decode("utf-8", "replace")
                    else:
                        body = request.body

                    try:
                        request_info["data"] = sanitize_data(json.loads(body))
                    except json.JSONDecodeError:
                        # But what is body? application/x-www-form-urlencoded I hope
                        request_info["data"] = sanitize_querystring(body)

                log_msg = f"OUTGOING (end): {request.method} {url}"

            if response is not None:
                log_msg += f" ({response.status_code})"

                response_info = {
                    "status_code": response.status_code,
                }

                response_text = getattr(response, "text", "")
                try:
                    # See if its json data
                    response_info["data"] = sanitize_data(json.loads(response_text))
                except json.JSONDecodeError:
                    response_info["data"] = sanitize_querystring(response_text)

    if exc_info:
        log_func = logger.error
    else:
        log_func = logger.info

    extra = {
        "start_time": start.strftime("%Y-%m-%d %H:%M:%S.%f"),
        "end_time": end and end.strftime("%Y-%m-%d %H:%M:%S.%f"),
        "response_time_ms": calculate_response_time_ms(start, end),
        "request": request_info,
        "response": response_info,
        "notes": notes,
        "smart": True,
        "direction": direction,
        "edge": edge,
    }

    log_func(log_msg, exc_info=exc_info, extra=extra)


# Expects to Handle django request objects
def log_incoming_request_event(
    *,
    start,
    end,
    logger,
    request=None,
    response=None,
    request_data=None,
    exc_info=None,
    notes=None,
    msg=None,
    method=None,
    url=None,
    direction=None,  # only exists to match existing interface
    edge=RequestEdge.END,
):
    from .config import config

    request_info = {}
    response_info = {}

    log_msg = None
    direction = RequestDirection.INCOMING

    with SensitivePathRequestContext(request):
        if request:
            path = sanitize_url(request.path)
            request_info = {
                "method": request.method,
                "remote_addr": request.META["REMOTE_ADDR"],
                "url_scheme": request.scheme,
                "path": path,
                "POST": sanitize_data(request.POST),
                "GET": sanitize_data(request.GET),
                "data": sanitize_data(request_data),
                "headers": sanitize_data(
                    {
                        h: sanitize_url(v) if h == "HTTP_REFERER" else v
                        for h, v in request.META.items()
                        if h.startswith("HTTP_")
                    },
                ),
            }

        if edge == RequestEdge.START:
            if request:
                log_msg = f"INCOMING (start): {request.method} {path}"
            else:
                # Better to provide a request on a START flow
                url = sanitize_url(url)
                log_msg = f"INCOMING (start): {method} {url}"

        else:
            # assert edge == RequestEdge.END
            if request:
                log_msg = f"INCOMING (end): {request.method} {request.path}"

            if response is not None:
                log_msg += f" ({response.status_code})"

                response_info = {
                    "status_code": response.status_code,
                }

                # Set _log_data to False on any Django response object before your
                # view returns it to prevent the response being logged.
                if getattr(response, "_log_data", True):
                    response_headers = getattr(response, "headers", {})
                    is_json_response = (
                        response_headers.get("Content-Type", "") == "application/json"
                    )

                    # We only ever log json responses
                    if config.LOG_RESPONSE_CONTENT and is_json_response:
                        response_info["data"] = sanitize_data(
                            json.loads(response.content)
                        )
                else:
                    response_info["data"] = {"NOT_LOGGED": "_log_data == False"}

    if exc_info:
        log_func = logger.error
    else:
        log_func = logger.info

    extra = {
        "start_time": start.strftime("%Y-%m-%d %H:%M:%S.%f"),
        "end_time": end and end.strftime("%Y-%m-%d %H:%M:%S.%f"),
        "response_time_ms": calculate_response_time_ms(start, end),
        "request": request_info,
        "response": response_info,
        "notes": notes,
        "smart": True,
        "direction": direction,
        "edge": edge,
    }

    log_func(log_msg, exc_info=exc_info, extra=extra)



import os

from setuptools import setup

CURDIR = os.path.abspath(os.path.dirname(__file__))
PACKAGE_PATH = "boston_logger"

with open(os.path.join(CURDIR, "README.md")) as readme:
    README = readme.read()

test_requires = [
    "pytest",
    "pytest-coverage",
    "pytest-mock",
    "python-dateutil",
]

TOX_ENV = os.environ.get("TOX_ENV_NAME", "django22")  # Added in tox 3.4
if "django20" in TOX_ENV:
    test_requires.append("Django>=2.0,<2.1")
elif "django21" in TOX_ENV:
    test_requires.append("Django>=2.1,<2.2")
elif "django30" in TOX_ENV:
    test_requires.append("Django>=3.0,<3.1")
elif "django31" in TOX_ENV:
    test_requires.append("Django>=3.1,<3.2")
elif "django32" in TOX_ENV:
    test_requires.append("Django>=3.2,<3.3")
elif "django40" in TOX_ENV:
    test_requires.append("Django>=4.0,<4.1")
elif "django41" in TOX_ENV:
    test_requires.append("Django>=4.1,<4.2")
else:
    # Default to the regular requirements
    test_requires.append("Django>=2.2,<5")


setup(
    name="boston-logger",
    packages=[PACKAGE_PATH],
    author="Aaron McMillin",
    author_email="AMcMillin@jbssolutions.com",
    long_description=README,
    long_description_content_type="text/markdown",
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: System :: Logging",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires=">=3.7, <4",
    install_requires=[
        "requests >= 2.18.0",
        "configular >= 1.1.0",
    ],
    extras_require={
        "django": ["Django >= 1.10"],
        "test": test_requires,
    },
    project_urls={"Source": "https://github.com/JBSinc/boston-logger"},
    use_scm_version={
        # PyPi doesn't allow local versions
        "local_scheme": "no-local-version",
    },
    setup_requires=["setuptools_scm"],
)



from configular import Settings
from configular.environ_loader import EnvironLoader

_defaults = {
    "ENABLE_OUTBOUND_REQUEST_LOGGING": True,
    "ENABLE_LOGGING_MIDDLEWARE": True,
    "ENABLE_SENSITIVE_PATHS_PROCESSOR": False,
    "ENABLE_REQUESTS_LOGGING": False,
    "MAX_VERBOSE_OUTPUT_LENGTH": 500,
    "MAX_JSON_DATA_TO_LOG": 0,  # Do not limit json output, by default
    "MIDDLEWARE_BLOCKLIST": ["admin:index", "swagger-docs"],
    "LOGGER_NAME": "boston_logger",
    "LOG_RESPONSE_CONTENT": False,
    "PREFER_TEXT_FALLBACK_MASKING": False,
    "SHOW_NESTED_KEYS_IN_SENSITIVE_PATHS": False,
}


class BlSettings(Settings):
    @property
    def request_logging_enabled(self):
        try:
            return str(self.ENABLE_REQUESTS_LOGGING).lower()[0] in ["y", "t"]
        except Exception:
            return False

    def reconfigure(self, *args, **kwargs):
        super().reconfigure(*args, **kwargs)

        if self.request_logging_enabled:
            from . import requests_monkey_patch  # noqa: F401

        if not isinstance(self.MIDDLEWARE_BLOCKLIST, list):
            raise ValueError("MIDDLEWARE_BLOCKLIST must be a list.")


config = BlSettings(defaults=_defaults, prefix="BOSTON_LOGGER", loaders=[EnvironLoader])



import json
import logging
from datetime import date, datetime
from decimal import Decimal

import pytest

from boston_logger.config import config
from boston_logger.logger import JsonFormatter, ObjectTypeEncoder


def test_format_limit():
    long_msg = "TEST " * 100

    config.MAX_JSON_DATA_TO_LOG = 500

    record = logging.makeLogRecord(
        {
            "msg": long_msg,
            "request": long_msg,
            "smart": True,
            "response": {"data": long_msg},
        }
    )
    formatter = JsonFormatter()

    fmt = formatter.format(record)
    fmt = json.loads(fmt)
    assert fmt["msg"] == long_msg
    assert fmt["request"] == long_msg

    # But because of the min 50, nothing get truncated
    config.MAX_JSON_DATA_TO_LOG = 5

    fmt = formatter.format(record)
    fmt = json.loads(fmt)
    assert fmt["max_data_exceeded"]

    # Trigger response truncation
    config.MAX_JSON_DATA_TO_LOG = 60

    fmt = formatter.format(record)
    fmt = json.loads(fmt)
    assert fmt["max_data_exceeded"]
    assert fmt["response"]["data"].endswith("**TRUNCATED**")

    # Unlimited
    config.MAX_JSON_DATA_TO_LOG = 0

    fmt = formatter.format(record)
    fmt = json.loads(fmt)
    assert fmt["msg"] == long_msg
    assert fmt["request"] == long_msg


@pytest.mark.parametrize(
    "obj,result",
    [
        (
            {2, 3, 1},
            {
                "value": [1, 2, 3],
                "type": "set",
            },
        ),
        (
            datetime(2021, 3, 4, 2, 30, 22),
            {
                "value": str(datetime(2021, 3, 4, 2, 30, 22)),
                "type": "datetime",
            },
        ),
        (
            date(2021, 3, 4),
            {
                "value": str(date(2021, 3, 4)),
                "type": "date",
            },
        ),
        (
            Decimal(1.5),
            {
                "value": str(Decimal(1.5)),
                "type": "Decimal",
            },
        ),
        (
            "String",
            {
                "value": "String",
                "type": None,
            },
        ),
        (
            3,
            {
                "value": 3,
                "type": None,
            },
        ),
    ],
)
def test_json_encoder(obj, result):
    resp = json.loads(json.dumps({"field": obj}, cls=ObjectTypeEncoder))
    if result["type"]:
        assert resp["field"]["value"] == result["value"]
        assert resp["field"]["type"] == result["type"]
    else:
        # Simple fields Don't get wrapped in a typing object
        assert resp["field"] == result["value"]
